# m209 : The base Class of M209 

import string
import sys
import random
import math
from   bib_util import *

class Wheel:
    def __init__(self,alphabet,shift):
        self.alphabet = alphabet
        self.lg = len(alphabet)
        self.pins = []
        for i in range(self.lg):
            self.pins.append( 0 )
        self.position = 0
        self.shiftWindow = shift

    def setPinsON(self,car):
        i = string.index(self.alphabet,car)
        self.pins[i] = 1

    def setPinsOFF(self,car):
        i = string.index(self.alphabet,car)
        self.pins[i] = 0
        
    def getAPins(self,car):
        i = string.index(self.alphabet,car)
        return self.pins[i]

    def getShift(self):
        return self.shiftWindow

    def turnClockwise(self):
        self.position = (self.position + 1 ) % self.lg

    def turnCounterClockwise(self):
        self.position = (self.position + self.lg - 1 ) % self.lg 

    def getLg( self):
        return len( self.alphabet )

    def getPosition(self):
        return self.position

    def setPosition(self, newPosition ):
        self.position = newPosition

    def getExternalKey(self):
        return self.alphabet[ self.position ]

    def getExternalShiftedKey(self):
        return self.alphabet[(self.position + self.shiftWindow)%self.lg]

    def getInternalKey(self):
        ch = ""
        for i in range(self.lg):
            if self.pins[i] :
                ch = ch + self.alphabet[i]
            else:
                ch = ch + "_"
        return ch

    def isActive(self):
        pos_active = ( self.position + self.shiftWindow ) % self.lg
        if self.pins[ pos_active ] != 0:
            return 1
        else:
            return 0

    def zero(self):
        for i in range(self.lg):
            self.pins[i] = 0

    def info(self):
        print "Lenght       : ", self.lg
        print "Alphabet     : ", self.alphabet
        print "Shitf        : ", self.shiftWindow
        print "Pins         : ", self.pins
        print "Internal Key : ", self.getInternalKey()
        print "External key : ", self.getExternalKey()
        print "Position: ", self.position, " Actif : ", self.isActive()

class Hagelin:
    def __init__(self, lettre_decalage, debug):
        self.debug = debug
        self.decalage = string.index(string.ascii_uppercase,lettre_decalage)
        self.wheels = []
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTUVWXYZ",15 ) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTUVXYZ",14) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTUVX",13) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTU",12) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRS",11) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQ",10  ) )
        self.nbOfWheels = len( self.wheels )
        self.lugs = [0]*6
        self.over = [0]*15
        self.counter  = 0

    def setSlide(self, lettre_decalage ):
        self.decalage = string.index(string.ascii_uppercase,lettre_decalage)

    def setDebug(self,debug):
        self.debug = debug

    def getDebug(self):
        return self.debug

    def setZeroOver(self):
        for i in range(len(self.over)):
            self.over[i]=0

    def setOverByKey(self,clef, valeur):
        overlaps = { '1-2':0, '1-3':1, '1-4':2, '1-5':3, '1-6':4, '2-3':5,
                     '2-4':6, '2-5':7, '2-6':8, '3-4':9, '3-5':10,'3-6':11,
                     '4-5':12,'4-6':13,'5-6':14 }
        indx = overlaps[clef] 
        self.over[indx] = valeur

    def setExternalKey( self, key ):
        for i in range( len(key) ):
            non_trouve = 1
            lettre_cle = key[i]
            lg = self.wheels[i].getLg()
            for j in range (lg):
                if lettre_cle == self.wheels[i].getExternalKey():
                    non_trouve = 0                  
                    break
                self.stepAWheel(i)
            if non_trouve: 
                print "**** Error: invalid external key  **** "; exit

    def setOver(self,lesOvers):
        self.over = lesOvers

    def getOver(self):
        return self.over

    def setLugs(self,lesLugs):
        self.lugs = lesLugs

    def getLugs(self):
        return self.lugs

    def setPinsON(self,laroue,char):
        self.wheels[laroue].setPinsON(char)

    def setPinsOFF(self,laroue,char):
        self.wheels[laroue].setPinsOFF(char)

    def setAllPins(self,laroue,ch):
        for i in range(len(ch)):
            if ch[i] == "1":
                self.setPinsON(laroue,
                self.wheels[laroue].alphabet[i] )
            else:
                self.setPinsOFF(laroue,
                self.wheels[laroue].alphabet[i] )

    def setAllPinsAllW(self,ch):
        self.setAllPins(0, ch[:26] )
        self.setAllPins(1, ch[26:51] )
        self.setAllPins(2, ch[51:74] )
        self.setAllPins(3, ch[74:95] )
        self.setAllPins(4, ch[95:114] )
        self.setAllPins(5, ch[114:] )

    def getPins(self,laroue,char):
        return self.wheels[laroue].getAPins(char)

    def stepAWheel(self, aWheel ):
        self.wheels[ aWheel ].turnClockwise()

    def step(self):
        for i in range( self.nbOfWheels ):      
            self.wheels[i].turnClockwise()
        self.counter = self.counter + 1

    def getPosition( self ):
        return self.counter

    def getPinsOfAWheel( self, aWheel ):
        return self.wheels[ aWheel ].getInternalKey()

    def stepBack(self):
        for i in range( self.nbOfWheels ):      
            self.wheels[i].turnCounterClockwise()
        self.counter = self.counter - 1
        if ( self.counter < 0 ):
            self.counter = 999

    def rewind(self):
        while ( self.counter != 0 ):
            self.stepBack()

    def getExternalKey( self ):
        ch = ""
        for i in range( self.nbOfWheels ):
            ch = ch + self.wheels[i].getExternalKey()
        return ch

    def cipher(self,char):
        shift = 0
        index = string.index(string.ascii_uppercase,char)
        activesPins = []
        for j in range( self.nbOfWheels ):
            if self.wheels[ j ].isActive():
                activesPins.append( 1 )
            else:
                activesPins.append( 0 )
        for j in range( self.nbOfWheels ):
            shift = shift + activesPins[j] * self.lugs[j]
        o = self.over
        shift = shift - o[0] * activesPins[0] * activesPins[1]
        shift = shift - o[1] * activesPins[0] * activesPins[2]
        shift = shift - o[2] * activesPins[0] * activesPins[3]
        shift = shift - o[3] * activesPins[0] * activesPins[4]
        shift = shift - o[4] * activesPins[0] * activesPins[5]
        shift = shift - o[5] * activesPins[1] * activesPins[2]
        shift = shift - o[6] * activesPins[1] * activesPins[3]
        shift = shift - o[7] * activesPins[1] * activesPins[4]
        shift = shift - o[8] * activesPins[1] * activesPins[5]
        shift = shift - o[9] * activesPins[2] * activesPins[3]
        shift = shift - o[10] * activesPins[2] * activesPins[4]
        shift = shift - o[11] * activesPins[2] * activesPins[5] 
        shift = shift - o[12] * activesPins[3] * activesPins[4]
        shift = shift - o[13] * activesPins[3] * activesPins[5]
        shift = shift - o[14] * activesPins[4] * activesPins[5]
        j = ( 26 + self.decalage - index + shift ) % 26 
        char2 = string.ascii_uppercase[ j ]
        if ( self.debug ):
            print "Counter: %03d Key: " % (self.getPosition()), 
            print "External Key: %6s " % (self.getExternalKey()),
            for i in range( self.nbOfWheels ):
                print "%d" % (activesPins[i]),
            print "%2d %c %c" % (shift, char, char2)
        self.step()
        return char2

    def getNbOfWheels(self):
        return self.nbOfWheels

    def toggleDebug( self ):
        if self.debug: self.debug = 0
        else: self.debug = 1

    def info(self):
        print "Slide   : ", self.decalage
        print "Lugs    : ", self.lugs
        print "Overlaps: ", self.over
        print "Actives Pins : ", 
        for i in range( self.nbOfWheels ):
            self.wheels[i].info()
#================================================
def chiffre( e, msg ):
        crypto = ""
        cpt = 0
        for i in range(len(msg)):
                car = e.cipher( msg[i] )
                crypto = crypto + car
                cpt = cpt + 1
        return crypto
#================================================
def restore( c, nomFichier ):
    if nomFichier == "":
        nomFichier = "config"
    f = open( nomFichier, "r" )
    for i in range(6):
        ch = f.readline()
        ch = ch.rstrip()
        lg = len(ch)
        for j in range(lg):
            if ch[j] != '_':
                c.setPinsON(i, ch[j] )
    ch = f.readline()
    ch = ch[0]
    #c.setLetterShift( ch )
    c.setSlide( ch )

    ch = f.readline()
    ch = ch.rstrip()
    lesLugs = ch.split(":")
    lesLugs = map(int,lesLugs)
    c.setLugs( lesLugs )

    ch = f.readline()
    nblignes = int (ch )
    for i in range( nblignes ):
        ch = f.readline()
        ch = ch.rstrip()
        (over_clef, over_nb) = ch.split(":")
        over_nb = int(over_nb)
        c.setOverByKey(over_clef, over_nb)
    f.close()
# ==== on fait juste un teste de chiffrement
if __name__ == "__main__":
    e = Hagelin ('Z',0)
#          01234567890123456789012345
    Pins="11010001101011000011011000"+\
         "0001101001110010011010100"+\
         "11000011010111000111111"+\
         "001011011000110100111"+\
         "0101110110001101001"+\
         "11010001001001101"
    e.setAllPinsAllW(Pins)
    Lugs="1:2:3:4:8:11"
    e.setLugs(map(int,Lugs.split(":")))
# overlaps: #4-5:1 #5-6:1
    Overs="0:0:0:0:0:0:0:0:0:0:0:0:1:0:1"
    e.setOver(map(int,Overs.split(":")))
    print chiffre(e,"AAAAAAAAAAAAAAAAAAAAAAAAAA")





# HC_PINS_Sullivan.py - Hill Climbing - search Pins - Lugs are known
# Algorith: the Sullivan's one

import string
import sys
import random
import math
import HC_SCORE
from bib_util import *
import m209
import getopt
import time

theOverlaps = [
 [0, 0], [0, 1], [0, 2], [0, 3], [0, 4],
 [1, 0], [1, 5], [1, 6], [1, 7], [0, 8],
 [2, 1], [2, 5], [2, 9], [2,10], [2,11],
 [3, 2], [3, 6], [3, 9], [3,12], [3,13],
 [4, 3], [4, 7], [4,10], [4,12], [4,14],
 [5, 4], [5, 8], [5,11], [5,13], [5,14]
]

#======================================
def initKeyPin( nbActivePins ):
    currentKeyPin= ""
    pins_array = [1] * nbActivePins
    pins_array.extend( [0] * (131 - nbActivePins) )
    random.shuffle( pins_array )
    pins_array = map( str, pins_array)
    return ''.join( pins_array )
    
#======================================
def usage():
	print "usage:"
	print "  -c crypto	The cryptogram"
	print "  -M method	The method (IC, UNI, BIG,TRI,...), default: IC"
	print "  -m file_method	The file used with method"
	print "  -l letter	The letter (shift, default: Z"
	print "  -L lugs	The lugs (ex: 1:5:4:13:2:7)"
	print "  -O overlaps	The overlaps (ex: 0:0:1...:1:1)"
	print "  -t tries 	Number of tries, default: 125"
#======================================

#--------------------------------------------------
try:
  opts, args = getopt.getopt(sys.argv[1:], 
	"ho:c:M:m:t:l:L:O:v", ["help", "output=", "crypto=","method=",
	"method_file=", "tries=", "letter=", "lugs=", "overlaps=", "verbose"])
except getopt.GetoptError as err:
  # print help information and exit:
  print sys.argv[0],": ", str(err) # option -a not recognized"
  usage()
  sys.exit(2)

output = None
verbose = False
MAX_Z = 125
METHOD = "IC"
file_method = ""
msg = ""
PINS = ""
LUGS = []
OVER = [0]*15
cle_letter_shift = "Z"

for o, a in opts:
  if o == "-v":
    verbose = True
  elif o in ("-h", "--help"):
    usage()
    sys.exit()
  elif o in ("-o", "--output"):
    output = a
  elif o in ("-c", "--crypto"):
    msg = litfic( a )
  elif o in ("-M", "--method"):
    METHOD = a
  elif o in ("-m", "--method_file"):
    file_method = a
  elif o in ("-t", "--tries"):
    MAX_Z = int(a)
  elif o in ("-l", "--letter"):
    cle_letter_shift = a
  elif o in ("-L", "--lugs"):
    LUGS=a.split(":")
    cle_lug=list( map(int,LUGS) )
  elif o in ("-O", "--overlaps="):
    tb = a.split(":")
    OVER = map(int,tb)
  else:
    assert False, "unhandled option"
if msg == "" or file_method == "":
  usage()
  sys.exit(1)
HC_SCORE.score_init[METHOD]( file_method )
#--------------------------------------------------
e = m209.Hagelin ( cle_letter_shift, 0 )
MAX_SCORE = 0
start_time = time.time()
nod = 0 # number of decrypts
currentKeyLug = cle_lug
oldKeyLug = currentKeyLug
currentKeyOve = OVER
oldKeyOve = currentKeyOve 
e.setLugs( currentKeyLug )
e.setOver( currentKeyOve )

# 1. Randomly set 40% of the pins in ARRAY to the active state.
# 2. Loop while the number of active pins in ARRAY is less than 60%
#    Remark: 40% == 52 pins, 60% == 80 pins
for nbActivePins in range(52,81):
  print "====> Number of Active Pins: ", nbActivePins
  # 3. Decrypt the message, using ARRAY and obtain a score for this 
  #    pin pattern
  # 3.1. Set the external wheels settings to AAAAAA
  e.rewind()

  # 3.2. Generate initial random pins 
  ARRAY = initKeyPin(nbActivePins)

  # 3.3 Save a copy of ARRAY in SAVE
  SAVE = ARRAY

  # 3.4 Configure key 
  e.setAllPinsAllW( ARRAY )

  # 3.5 Decipher with current key
  plain = m209.chiffre( e, msg ) ; nod += 1

  # 3.6 Calculate fitness 
  HIGH_SCORE = HC_SCORE.score[METHOD]( plain )
  MAX_CLAIR = plain

  # 4. Swap an score all pin pairs in ARRAY. Retain only the swap that
  #    increase the score. Record any higher score in a log file.
  # 5. Loop for typically 125 loops.
  for i in range( MAX_Z ): 
    k1 = 0 ; k2 = 1
    while k1 < 129 or k2 < 130:
      # --- External Key sets to "AAAAAA"
      e.rewind()

      # --- Swap a pair of pins
      tb = list( SAVE )
      p1 = k1
      p2 = k2
      k2 = k2 + 1 
      if k2 == 130 : k1 = k1 + 1 ; k2 = k1 + 1
      if tb[p1] == tb[p2]: continue
      x = tb[p1]
      tb[p1] = tb[p2]
      tb[p2] = x
      ARRAY = "".join( tb )
      e.setAllPinsAllW( ARRAY )

      # --- To cipher
      plain = m209.chiffre( e, msg ) ; nod += 1

      # --- Calculate fitness
      score = HC_SCORE.score[METHOD]( plain )

      # --- Keep it if best
      if score > HIGH_SCORE :
        HIGH_SCORE = score			
        SAVE = ARRAY
        MAX_CLAIR = str(plain)

      # ... Record any higher score in a log file.
      if HIGH_SCORE > MAX_SCORE :
        elaps = time.time() - start_time
        MAX_SCORE = HIGH_SCORE
        print ">>>MAX SCORE: ", HIGH_SCORE, "==", MAX_CLAIR[0:50]
        print "CLE: ", SAVE, oldKeyLug, oldKeyOve, nod, elaps, 
        print nbActivePins, i, k1, k2

    # 6. Introduce some small random variation in ARRAY, keeping
    #    the number of active pins constant.
    for j in range(0):
      tb = list( SAVE )
      i1 = random.randint(0,130)
      i2 = random.randint(0,130)
      if i1 == i2 : continue    
      if tb[i1] == "1" : x = "0"
      else: x = "1"
      tb[ i1 ] = x
      if tb[i2] == "1" : x = "0"
      else: x = "1"
      tb[ i2 ] = x
      SAVE = "".join( tb )




# Hill Climbing - M-209 - Cipher text Attack
# HC_SULLIVAN.py  v0.9 : 10/09/15 JFB 
# based on Geoff Sullivan Algorithm - Cryptologia (2002) - 

import string
import sys
import random
import math
import HC_SCORE
import m209
from HC_modif_KPC import *
from bib_util import *
import getopt
import time
import copy

#======================================
def usage():
    print "usage:"
    print "  -c crypto	The cryptogram"
    print "  -M method	The method (BIG,TRI,...), default: IC"
    print "  -m file_method	The file used with method"
    print "  -l letter	The letter (shift, default: Z"
    print "  -P pins	The Pins"
    print "  -L lugs	The lugs (ex: 1:5:4:13:2:7)"
    print "  -S list	Skip P(ins), S(wap), L(ugs)"
    print "  -B         I choose the Best solution from first stage"
#======================================

WheelsPositions = [
    [  0, 26 ],
    [ 26, 25 ],
    [ 51, 23 ],
    [ 74, 21 ],
    [ 95, 19 ],
    [114, 17 ]
] 
TwoWheelsKick = [
  # r1,k1,r2,k2,   x, xl,   y, ylg
  [  0, 1, 1, 2 ],   # 0  wheels 1 (26) and 2 (25)
  [  0, 2, 1, 1 ],   # 1
  [  0, 1, 2, 2 ],   # 2  wheels 1 (26) and 3 (23)
  [  0, 2, 2, 1 ],   # 3
  [  0, 1, 3, 2 ],   # 4  wheels 1 (26) and 4 (21)
  [  0, 2, 3, 1 ],   # 5
  [  0, 1, 4, 2 ],   # 6  wheels 1 (26) and 5 (19)
  [  0, 2, 4, 1 ],   # 7
  [  0, 1, 5, 2 ],   # 8  wheels 1 (26) and 6 (17)
  [  0, 2, 5, 1 ],   # 9  
  [  1, 1, 2, 2 ],   #10  wheels 2 (25) and 3 (23)
  [  1, 2, 2, 1 ],   #11
  [  1, 1, 3, 2 ],   #12  wheels 2 (25) and 4 (21)
  [  1, 2, 3, 1 ],   #13
  [  1, 1, 4, 2 ],   #14  wheels 2 (25) and 5 (19)
  [  1, 2, 4, 1 ],   #15
  [  1, 1, 5, 2 ],   #16  wheels 2 (25) and 6 (17)
  [  1, 2, 5, 1 ],   #17
  [  2, 1, 3, 2 ],   #18  wheels 3 (23) and 4 (21)
  [  2, 2, 3, 1 ],   #19
  [  2, 1, 4, 2 ],   #20  wheels 3 (23) and 5 (19)
  [  2, 2, 4, 1 ],   #21
  [  2, 1, 5, 2 ],   #22  wheels 3 (23) and 6 (17)
  [  2, 2, 5, 1 ],   #23
  [  3, 1, 4, 2 ],   #24  wheels 4 (23) and 5 (21)
  [  3, 2, 4, 1 ],   #25
  [  3, 1, 5, 2 ],   #26  wheels 4 (23) and 6 (17)
  [  3, 2, 5, 1 ],   #27
  [  4, 1, 5, 2 ],   #28  wheels 5 (21) and 6 (17)
  [  4, 2, 5, 1 ]    #29  
]
#--------------------------------------------------
try:
  opts, args = getopt.getopt(sys.argv[1:], 
	"ho:c:M:m:l:L:S:P:vB", ["help", "output=", "crypto=","method=",
	"method_file=", "letter=", "lugs=", "skip=","pins=", "verbose", "best"])
except getopt.GetoptError as err:
  # print help information and exit:
  print sys.argv[0],": ", str(err) # option -a not recognized"
  usage()
  sys.exit(2)

output = None
verbose = False
best = False
MAX_Z = 15
METHOD = "IC"
file_method = "IC"
msg = ""
PINS = ""
LUGS = []
OVER = [0]*16
SKIP_LUGS = False
SKIP_PINS = False
SKIP_SWAP = False
cle_letter_shift = "Z"

for o, a in opts:
  if o == "-v":
    verbose = True
  elif o in ("-B", "--best"):
    best = True
  elif o in ("-h", "--help"):
    usage()
    sys.exit()
  elif o in ("-o", "--output"):
    output = a
  elif o in ("-c", "--crypto"):
    msg = litfic( a )
  elif o in ("-M", "--method"):
    METHOD = a
  elif o in ("-m", "--method_file"):
    file_method = a
  elif o in ("-l", "--letter"):
    cle_letter_shift = a
  elif o in ("-L", "--lugs"):
    LUGS=a.split(":")
    cle_lug=list( map(int,LUGS) )
  elif o in ("-S", "--skip"):
    les_algo = list(a)
    if 'L' in les_algo: SKIP_LUGS = True
    if 'P' in les_algo: SKIP_PINS = True
    if 'W' in les_algo: SKIP_SWAP = True
  elif o in ("-P", "--pins="):
    PINS = a
  else:
    assert False, "unhandled option"
if msg == "" or file_method == "":
  usage()
  sys.exit(1)
HC_SCORE.score_init[METHOD]( file_method )
#--------------------------------------------------
e = m209.Hagelin ( cle_letter_shift, 0 )
nod = 0 # Number of Decipherment 
start_time = time.time()
#------------------ first Stage -------------------
print "============ first Stage ================="
solutions = [0]*30
MAX_Z = 0 ; MAX_Z_SCORE = 0
for z in range(	30 ):
  # 1. Set the external wheels settings to AAAAAA
  e.rewind()
  MAX_SCORE = 0

  # 2. Generate initial random pins 
  #    (or fixes setting according to parameters)
  (wh1, kick1, wh2, kick2) = TwoWheelsKick[z]
  (w1st, w1lg) = WheelsPositions[ wh1 ]
  (w2st, w2lg) = WheelsPositions[ wh2 ]

  LUGS=[0]*6
  LUGS[wh1]=kick1
  LUGS[wh2]=kick2
  currentKeyLug = LUGS
  oldKeyLug = currentKeyLug

  currentKeyOve = [0] * 16
  oldKeyOve = currentKeyOve

  currentKeyPin = initKeyPin()
  tb = list(currentKeyPin)
  for i in range(131):
    if not ((i>=w1st and i<(w1st+w1lg))or(i>=w2st and i<(w2st+w2lg))):
      tb[i] = "0"
  currentKeyPin = "".join( tb )
  oldKeyPin = currentKeyPin

  #===> Configurer la cle
  e.setAllPinsAllW( currentKeyPin )
  e.setLugs( currentKeyLug )
  e.setOver( currentKeyOve )

  # --- 2.1 On dechiffre avec la cle courante
  plain = m209.chiffre( e, msg ); nod +=1

  # --- 2.2 on calcule le fitness
  HIGH_SCORE = HC_SCORE.score[METHOD]( plain )
  MAX_CLAIR = plain

  # 3. Repeat the following steps until no longer improves
  anyImprovement = True
  while anyImprovement:
    anyImprovement = False

    # 3a. Check all pins transformation as long as an improvement
    if not SKIP_PINS:
     improved = True
     while improved:
      improved = False
      e.setLugs( oldKeyLug )
      e.setOver( oldKeyOve )

      # ---- 
      for iw1w2 in range(w1lg+w2lg): 
        # --- External Key sets to "AAAAAA"
        e.rewind()

        # --- Toogle a Pin
        if iw1w2 < w1lg :  i = w1st + iw1w2
        else:              i = w2st + iw1w2 - w1lg
        tb = list( oldKeyPin )
        if tb[i] == "1" : x = "0"
        else: x = "1"
        tb[ i ] = x
        currentKeyPin = "".join( tb )
        e.setAllPinsAllW( currentKeyPin )

        # --- To cipher
        plain = m209.chiffre( e, msg ) ; nod += 1

        # --- Calculate fitness
        score = HC_SCORE.score[METHOD]( plain )

        # --- Keep it if best
        if score > HIGH_SCORE :
          HIGH_SCORE = score			
          oldKeyPin = str(currentKeyPin)
          MAX_CLAIR = str(plain)
          anyImprovement = improved = True

     if not SKIP_SWAP:
       improved = True
       while improved:
         improved = False
         e.setLugs( oldKeyLug )
         e.setOver( oldKeyOve )
         k1 = 0 ; k2 = 1
         kMax = w1lg + w2lg
         #while k1 < 129 or k2 < 130:
         while k1 < (kMax - 2) or k2 < (kMax - 1):
           # --- External Key sets to "AAAAAA"
           e.rewind()

           # --- Swap a pair of pins
           tb = list( oldKeyPin )
           if (k1 < w1lg):
             p1 = k1 + w1st
           else:
             p1 = k1 + w2st - w1lg
           if (k2 < w1lg):
             p2 = k2 + w1st
           else:
             p2 = k2 + w2st - w1lg
           k2 = k2 + 1 
           if k2 == (kMax - 1) : k1 = k1 + 1 ; k2 = k1 + 1
           if tb[p1] == tb[p2]: continue
           x = tb[p1]
           tb[p1] = tb[p2]
           tb[p2] = x
           currentKeyPin = "".join( tb )
           e.setAllPinsAllW( currentKeyPin )

           # --- To cipher
           plain = m209.chiffre( e, msg ) ; nod += 1

           # --- Calculate fitness
           score = HC_SCORE.score[METHOD]( plain )

           # --- Keep it if best
           if score > HIGH_SCORE :
             HIGH_SCORE = score			
             oldKeyPin = str(currentKeyPin)
             MAX_CLAIR = str(plain)
             anyImprovement = improved = True

  if HIGH_SCORE > MAX_SCORE :
    elasp = time.time() - start_time
    MAX_SCORE = HIGH_SCORE
    print z, ">>>MAX_SCORE: ", HIGH_SCORE, "==", MAX_CLAIR[0:50],
    print "CLE: ", oldKeyPin, oldKeyLug, oldKeyOve, "NoD: ",  nod , elasp
    sys.stdout.flush()
  if HIGH_SCORE > MAX_Z_SCORE:
    MAX_Z = z ; MAX_Z_SCORE = HIGH_SCORE
  copyOldKeyLug = copy.copy(oldKeyLug)
  copyOldKeyPin = copy.copy(oldKeyPin)
  solutions[z] = [ HIGH_SCORE, MAX_CLAIR[0:50], copyOldKeyPin, copyOldKeyLug ]

print "Best of Stage1: ", MAX_Z, ",", solutions[MAX_Z]
sys.stdout.flush()
print ">>>>>>>>>>>>>>> second stage <<<<<<<<<<<<<<<<<"
MAXMAX=0 
if best:
    MIN_Z = MAX_Z ; MAX_Z +=1
else:
    MIN_Z = 0 ; MAX_Z = 30
for z in range(MIN_Z,MAX_Z):
 print "z:", z, " ", solutions[z]
 #---------------- 
 WheelsToScan = [0,1,2,3,4,5]
 w1 = TwoWheelsKick[z][0]
 WheelsToScan.remove(w1)
 w2 = TwoWheelsKick[z][2]
 WheelsToScan.remove(w2)
 MAX_SCORE = 0
 MAX_CLAIR = ""
 while len(WheelsToScan) > 0:
  print "Wheels To Scan: ", WheelsToScan
  maxWheel = WheelsToScan[0]
  for w in range(len(WheelsToScan)):
   for nLugs in range(2,14):
    aW = WheelsToScan[w]
    oldKeyPin = copy.copy(solutions[z][2])
    oldKeyLug = copy.copy(solutions[z][3])
    oldKeyLug[aW] = nLugs      
    HIGH_SCORE= solutions[z][0]
    tb = list(oldKeyPin)
    wst = WheelsPositions[aW][0]
    wlg = WheelsPositions[aW][1]
    for i in range(wlg):
      if random.random() > 0.5: tb[i+wst]="1"
      else: tb[i+wst]="0"
    currentKeyPin = "".join( tb )
    oldKeyPin = currentKeyPin

    anyImprovement = True
    while anyImprovement:
      anyImprovement = False

      # 3a. Check all pins transformation as long as an improvement
      if not SKIP_PINS:
       improved = True
       while improved:
        improved = False
        e.setLugs( oldKeyLug )
        e.setOver( oldKeyOve )

        # ---- 
        for i in range(wst, wst+wlg):
          # --- External Key sets to "AAAAAA"
          e.rewind()

          # --- Toogle a Pin
          tb = list( oldKeyPin )
          if tb[i] == "1" : x = "0"
          else: x = "1"
          tb[ i ] = x
          currentKeyPin = "".join( tb )
          e.setAllPinsAllW( currentKeyPin )

          # --- To cipher
          plain = m209.chiffre( e, msg ) ; nod += 1

          # --- Calculate fitness
          score = HC_SCORE.score[METHOD]( plain )

          # --- Keep it if best
          if score > HIGH_SCORE :
            HIGH_SCORE = score
            oldKeyPin = str(currentKeyPin)
            MAX_CLAIR = str(plain)
            anyImprovement = improved = True

      if not SKIP_SWAP:
       improved = True
       while improved:
         improved = False
         e.setLugs( oldKeyLug )
         e.setOver( oldKeyOve )
         k1 = 0 ; k2 = 1
         kMax = wlg 
         while k1 < (kMax - 2) or k2 < (kMax - 1):
           # --- External Key sets to "AAAAAA"
           e.rewind()

           # --- Swap a pair of pins
           tb = list( oldKeyPin )
           p1 = k1 + wst
           p2 = k2 + wst
           k2 = k2 + 1
           if k2 == (kMax - 1) : k1 = k1 + 1 ; k2 = k1 + 1
           if tb[p1] == tb[p2]: continue
           x = tb[p1]
           tb[p1] = tb[p2]
           tb[p2] = x
           currentKeyPin = "".join( tb )
           e.setAllPinsAllW( currentKeyPin )

           # --- To cipher
           plain = m209.chiffre( e, msg ) ; nod += 1

           # --- Calculate fitness
           score = HC_SCORE.score[METHOD]( plain )

           # --- Keep it if best
           if score > HIGH_SCORE :
             HIGH_SCORE = score
             oldKeyPin = str(currentKeyPin)
             MAX_CLAIR = str(plain)
             anyImprovement = improved = True

    if 1:
          elasp = time.time() - start_time
          print z, "===MAX_SCORE: ", HIGH_SCORE, "==", MAX_CLAIR[0:50],
          print "CLE: ", oldKeyPin, oldKeyLug, oldKeyOve, "NoD: ",  nod , elasp,
          print aW, WheelsToScan
          sys.stdout.flush()
          if HIGH_SCORE > MAX_SCORE :
            MAX_SCORE = HIGH_SCORE
            maxKeyLug = copy.copy( oldKeyLug )
            maxKeyPin = copy.copy( oldKeyPin )
            maxWheel  = aW
            MAXMAXCLR = copy.copy( MAX_CLAIR)
          if HIGH_SCORE > MAXMAX :
            MAXMAX = HIGH_SCORE
            maxmaxLug = copy.copy( oldKeyLug )
            maxmaxPin = copy.copy( oldKeyPin )
            maxmaxCLR = copy.copy( MAX_CLAIR )

  print "-------Wheel: ", maxWheel + 1, " Lugs: ", maxKeyLug[ maxWheel ], 
  print MAX_SCORE, maxKeyLug, maxKeyPin
  sys.stdout.flush()
  solutions[z][2] = copy.copy( maxKeyPin )
  solutions[z][3] = copy.copy( maxKeyLug )
  WheelsToScan.remove( maxWheel )
  
print "========>FIN<=========="
elasp = time.time() - start_time
print MAXMAX, maxmaxLug, maxmaxPin, maxmaxCLR[0:50], elasp





# HC_SCORE.py : Les methodes de scores

import math
import string
#=========================================
def initIC( fic_IC ):
	1
#=========================================
def initPlain( fic_plain ):
	global lePlain
        f = open(fic_plain,"r")
        leTexte = f.readlines()
        texteConcat = ""
        f.close()
        for i in range(len(leTexte)):
                ch = leTexte[i]
                ch = ch.upper()
                for j in range(len(ch)):
                        if not is_maj(ch[j]):
                                continue
                        texteConcat = texteConcat + ch[j]
        lePlain = texteConcat
#===========Unigram=======================
def initUni( fic_uni ):
	global lesUni
	for i in range(26):
		lesUni.append( 0. )
	f_uni = open( "STATS/" + fic_uni + ".uni", "r")
	while 1:
		ch = f_uni.readline()
		if ch == "":break
		ch = ch.strip()
 		(uni,fre) = ch.split(',')
		i = string.index(string.ascii_uppercase, uni)
		#lesUni[i]= math.log(int(fre))
		lesUni[i]= float(fre)
	#for i in range(26):
	#	print lesUni[i],
	print
#===========Bigrammes=====================
def initBig( fic_big ):
	global lesBig
	for i in range(26*26):
        	lesBig.append( 1. )

	#f_big = open(english.big","r")
	f_big = open( "STATS/" + fic_big + ".big" , "r" )
	while 1:
	        ch = f_big.readline()
        	if ch == "": break
	        ch = ch.strip()
        	(big,fre) = ch.split(',')
	        #print "<",big, "=", fre, ">"
	        i1 = string.index(string.ascii_uppercase, big[0])
        	i2 = string.index(string.ascii_uppercase, big[1])
	        lesBig[ i1*26 + i2 ] = math.log(int(fre))
	#       lesBig[ i1*26 + i2 ] = int(fre)

	#for i in range(26*26):
	#       print lesBig[i],
#===============================================
def initTri( fic_tri ):
	global lesTri
        for i in range(26*26*26):
                lesTri.append( 1. )

        f_tri = open( "STATS/" + fic_tri + ".tri" ,"r")
        while 1:
                ch = f_tri.readline()
                if ch == "": break
                ch = ch.strip()
                (tri,fre) = ch.split(',')
                #print "<",tri, "=", fre, ">"
                i1 = string.index(string.ascii_uppercase, tri[0])
                i2 = string.index(string.ascii_uppercase, tri[1])
                i3 = string.index(string.ascii_uppercase, tri[2])
                lesTri[ i1*26*26 + i2*26 + i3 ] = math.log(int(fre))
#===============================================
def initQua( fic_qua ):
	global lesQua
        f_qua = open( "STATS/" + fic_qua + ".qua","r")
        while 1:
                ch = f_qua.readline()
                if ch == "": break
                ch = ch.strip()
                (qua,fre) = ch.split(',')
                #print "<",qua, "=", fre, ">"
                lesQua[qua]=math.log(int(fre))
#================================================
def is_maj(car):

        if string.find(string.ascii_uppercase,car) == -1:
                return 0
        else:
                return 1
# ===================================
def score_IC(texte):
        f = []
        for i in range(26):
                f.append(0)
        N = len(texte)
        for i in range(N):
                indice = string.index(string.ascii_uppercase, texte[i])
                f[indice] = f[indice] + 1
        total = 0
        for i in range(26):
                if ( f[i] > 1):
                        total = total + f[i]*(f[i]-1)
        return total / float(N * (N-1))
#=====================================
def score_plain(texte):
	global lePlain
	total = 0
	N = len(texte)
	for i in range(N):
		if texte[i] == lePlain[i]:
			total = total + 1
	return total / float(N)
#=====================================
def score_ADE(texte):
	global lePlain
	N = len(texte)
	maxds = N * 26
	for i in range(N):
		x = abs(ord(texte[i]) - ord(lePlain[i])) 
		if x > 13:
			x -= 13
		maxds = maxds - x 
	return maxds / float(N * 26)
#=====================================
def score_ADE2(texte):
    global lePlain
    N = len(texte)
    maxds = N * 26
    for i in range(N):
        x = abs(ord(texte[i]) - ord(lePlain[i]))
        if x > 13:
            x = abs( x - 26 )
        maxds -= x
    return maxds / float(N * 26 )
#=====================================
def score_uni(texte):
	global lesUni
	total = 0
	for i in range( len(texte) ):
		lettre = texte[i]
		idx = string.uppercase.index( lettre )
		total += lesUni[ idx ]
	return total
#=====================================
def score_big(texte):
	global lesBig
	total = 0
	for i in range( len(texte) - 1 ):
	        lettre = texte[i]
        	idx1 = string.uppercase.index( lettre )
	
        	lettre = texte[i+1]
        	idx2 = string.uppercase.index( lettre )

	        total += lesBig[ idx1*26 + idx2 ]
	return total
#=======================================
def score_tri(texte):
	global lesTri
        total = 0
        for i in range( len(texte) - 2 ):
                lettre = texte[i]
                idx1 = string.uppercase.index( lettre )

                lettre = texte[i+1]
                idx2 = string.uppercase.index( lettre )

                lettre = texte[i+2]
                idx3 = string.uppercase.index( lettre )

                total += lesTri[ idx1*26*26 + idx2*26 + idx3 ]
        return total
#=======================================
def score_qua(texte):
	global lesQua
        total = 0
        for i in range( len(texte) - 3 ):
                qua = texte[i] + texte[i+1] + texte[i+2] + texte[i+3]
                if lesQua.has_key(qua):
                        total += lesQua[qua]
        return total

#>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
lesUni = []
lesBig = []
lesTri = []
lesQua = {}

score_init = { 'IC':initIC, 'PLN':initPlain, 'UNI':initUni, 'BIG':initBig, 
               'TRI':initTri, 'QUA':initQua, 'ADE':initPlain, 'ADE2':initPlain
}
score = { 'IC':score_IC, 'PLN':score_plain, 'UNI':score_uni, 'BIG':score_big, 
          'TRI':score_tri, 'QUA':score_qua, 'ADE':score_ADE, 'ADE2':score_ADE2
}







# bib_util.py 

import string
#================================================
def is_maj(car):

        if string.find(string.ascii_uppercase,car) == -1:
                return 0
        else:
                return 1
# ===================================
def litfic( fic ):
        f = open(fic,"r")
        leTexte = f.readlines()
        texteConcat = ""
        f.close()
        for i in range(len(leTexte)):
                ch = leTexte[i]
                ch = ch.upper()
                for j in range(len(ch)):
                        if not is_maj(ch[j]):
                                continue
                        texteConcat = texteConcat + ch[j]
        return texteConcat
#======================================

